# -*- coding: utf-8 -*-
"""
Created on Wed Jun 30 19:29:40 2021

@author: vorst
"""

# Python imports

# Third party imports
import numpy as np
from sklearn.metrics import accuracy_score, precision_score, jaccard_score
from sklearn.model_selection import ShuffleSplit
from sklearn.dummy import DummyClassifier
from sklearn.neighbors import RadiusNeighborsClassifier
from sklearn.metrics import make_scorer

# Local imports
from bag_cross_validate import cross_validate_bag, BagScorer, bags_2_si

# Global definitions

#%%

# Create some dummy data
"""Generate some dummy data
Create bags and single-instance data
A set of bags have a shape [n x (m x p)], and can be through of as an
array of bag instances.
n is the number of bags
m is the number of instances within each bag (this can vary between bags)
p is the feature space of each instance"""
n_bags = 1000
m_instances = 8 # Static number of bags
p = 5
bags = []
# 25% negative class, 75% positive class
# Bags are created with random data, of shape (n, (m,p))
labels = np.concatenate((np.ones(int(n_bags*0.5)),
                         np.zeros(int(n_bags*(1-0.5))),
                         ))
bags = np.random.randint(low=0, high=2, size=(n_bags, m_instances, p))
print("This is what a bag looks like: \n{}".format(bags[0]))

# Split dummy dataset dataset
rs = ShuffleSplit(n_splits=1, test_size=0.2, train_size=0.8)
train_index, test_index = next(rs.split(bags, labels))
train_bags, train_labels = bags[train_index], labels[train_index]
test_bags, test_labels = bags[test_index], labels[test_index]
        
# Create an estimator
dumb = DummyClassifier(strategy='constant', constant=1)
radiusNeighbor = RadiusNeighborsClassifier(weights='distance', 
                                           algorithm='auto',
                                           p=1, # Manhattan distance
                                           )

# Create an evaluation metric
# Multiple evaluation metrics are allowed
accuracy_scorer = make_scorer(accuracy_score)
bagAccScorer = BagScorer(accuracy_scorer) # Accuracy score, no factory function
precision_scorer = make_scorer(precision_score, average='binary')
bagPreScorer = BagScorer(precision_scorer)
jaccard_scorer = make_scorer(jaccard_score, average='binary')
bagJacScorer = BagScorer(jaccard_scorer)
scoring = {'bag_accuracy':bagAccScorer,
           'bag_precision':bagPreScorer,
           'bag_jaccard':bagJacScorer,
           }


#%%

# Cross validate the dummy data and estimator
result_dumb = cross_validate_bag(estimator=dumb, 
                            X=train_bags, 
                            y=train_labels, 
                            groups=None, 
                            scoring=scoring, # Custom scorer... 
                            cv=2,
                            n_jobs=3, 
                            verbose=0, 
                            fit_params=None,
                            pre_dispatch='2*n_jobs', 
                            return_train_score=False,
                            return_estimator=False, 
                            error_score=np.nan)

result_neighbor = cross_validate_bag(estimator=radiusNeighbor, 
                            X=train_bags, 
                            y=train_labels, 
                            groups=None, 
                            scoring=scoring, # Custom scorer... 
                            cv=3,
                            n_jobs=2, 
                            verbose=0, 
                            fit_params=None,
                            pre_dispatch='2*n_jobs', 
                            return_train_score=False,
                            return_estimator=False, 
                            error_score=np.nan)

# Display the results
msg=("\nOur dummy estimator tried his best, and predicted {} percent of bags " 
    "correctly")
msg = msg.format(result_dumb['test_bag_accuracy'])
print(msg)

msg=("\nOur neighbor estimator didnt fair well either, and predicted {} percent "
     "of bags correctly")
msg = msg.format(result_neighbor['test_bag_accuracy'])
print(msg)


# -*- coding: utf-8 -*-
"""
Created on Wed Jun 24 20:53:13 2020

@author: z003vrzk
"""


# Python imports
from collections import Counter
import warnings
import numbers
import time
from traceback import format_exc
from typing import Union

# Sklearn imports
from sklearn.utils.validation import (_check_fit_params, 
                                      check_is_fitted,
                                      NotFittedError, 
                                      _num_samples,
                                      _deprecate_positional_args)
from sklearn.utils.metaestimators import _safe_split
from sklearn.model_selection._validation import (_score, 
                                                 _aggregate_score_dicts, 
                                                 logger)
from sklearn.base import (is_classifier, 
                          clone)
from sklearn.utils import indexable
from sklearn.metrics._scorer import _check_multimetric_scoring
from sklearn.exceptions import FitFailedWarning
from sklearn.model_selection._split import check_cv

# Third party imports
from scipy.sparse import vstack
from scipy.sparse import csr_matrix
from joblib import (Parallel, 
                    delayed)
import numpy as np

#%%

def _fit_and_score(estimator, X, y, scorer, train, test, verbose,
                   parameters, fit_params, return_train_score=False,
                   return_parameters=False, return_n_test_samples=False,
                   return_times=False, return_estimator=False,
                   split_progress=None, candidate_progress=None,
                   error_score=np.nan):

    """Fit estimator and compute scores for a given dataset split.
    Parameters
    ----------
    estimator: estimator object implementing 'fit'
        The object to use to fit the data.
    X: array-like of shape (n_samples, n_features)
        The data to fit.
    y: array-like of shape (n_samples,) or (n_samples, n_outputs) or None
        The target variable to try to predict in the case of
        supervised learning.
    scorer: A single callable or dict mapping scorer name to the callable
        If it is a single callable, the return value for ``train_scores`` and
        ``test_scores`` is a single float.
        For a dict, it should be one mapping the scorer name to the scorer
        callable object / function.
        The callable object / fn should have signature
        ``scorer(estimator, X, y)``.
    train: array-like of shape (n_train_samples,)
        Indices of training samples.
    test: array-like of shape (n_test_samples,)
        Indices of test samples.
    verbose: int
        The verbosity level.
    error_score: 'raise' or numeric, default=np.nan
        Value to assign to the score if an error occurs in estimator fitting.
        If set to 'raise', the error is raised.
        If a numeric value is given, FitFailedWarning is raised. This parameter
        does not affect the refit step, which will always raise the error.
    parameters: dict or None
        Parameters to be set on the estimator.
    fit_params: dict or None
        Parameters that will be passed to ``estimator.fit``.
    return_train_score: bool, default=False
        Compute and return score on training set.
    return_parameters: bool, default=False
        Return parameters that has been used for the estimator.
    split_progress: list or tuple, optional, default: None
        A list or tuple of format (<current_split_id>, <total_num_of_splits>)
    candidate_progress: list or tuple, optional, default: None
        A list or tuple of format
        (<current_candidate_id>, <total_number_of_candidates>)
    return_n_test_samples: bool, default=False
        Whether to return the ``n_test_samples``
    return_times: bool, default=False
        Whether to return the fit/score times.
    return_estimator: bool, default=False
        Whether to return the fitted estimator.
    Returns
    -------
    result: dict with the following attributes
        train_scores: dict of scorer name -> float
            Score on training set (for all the scorers),
            returned only if `return_train_score` is `True`.
        test_scores: dict of scorer name -> float
            Score on testing set (for all the scorers).
        n_test_samples: int
            Number of test samples.
        fit_time: float
            Time spent for fitting in seconds.
        score_time: float
            Time spent for scoring in seconds.
        parameters: dict or None
            The parameters that have been evaluated.
        estimator: estimator object
            The fitted estimator.
    """
    progress_msg = ""
    if verbose > 2:
        if split_progress is not None:
            progress_msg = f" {split_progress[0]+1}/{split_progress[1]}"
        if candidate_progress and verbose > 9:
            progress_msg += (f"; {candidate_progress[0]+1}/"
                             f"{candidate_progress[1]}")

    if verbose > 1:
        if parameters is None:
            params_msg = ''
        else:
            sorted_keys = sorted(parameters)  # Ensure deterministic o/p
            params_msg = (', '.join(f'{k}={parameters[k]}'
                                    for k in sorted_keys))
    if verbose > 9:
        start_msg = f"[CV{progress_msg}] START {params_msg}"
        print(f"{start_msg}{(80 - len(start_msg)) * '.'}")

    # Adjust length of sample weights
    fit_params = fit_params if fit_params is not None else {}
    fit_params = _check_fit_params(X, fit_params, train)

    if parameters is not None:
        # clone after setting parameters in case any parameters
        # are estimators (like pipeline steps)
        # because pipeline doesn't clone steps in fit
        cloned_parameters = {}
        for k, v in parameters.items():
            cloned_parameters[k] = clone(v, safe=False)

        estimator = estimator.set_params(**cloned_parameters)

    start_time = time.time()

    X_train, y_train = _safe_split(estimator, X, y, train)
    X_test, y_test = _safe_split(estimator, X, y, test, train)

    # Fit the estimator
    # Let the custom bag scorer handle fitting of the estimator
    result = {}
    try:
        estimator = _BagScorer_estimator_fit(estimator, 
                                             X_train, 
                                             y_train, 
                                             scorer, 
                                             **fit_params)

    except Exception as e:
        # Note fit time as time until error
        fit_time = time.time() - start_time
        score_time = 0.0
        if error_score == 'raise':
            raise
        elif isinstance(error_score, numbers.Number):
            if isinstance(scorer, dict):
                test_scores = {name: error_score for name in scorer}
                if return_train_score:
                    train_scores = test_scores.copy()
            else:
                test_scores = error_score
                if return_train_score:
                    train_scores = error_score
            warnings.warn("Estimator fit failed. The score on this train-test"
                          " partition for these parameters will be set to %f. "
                          "Details: \n%s" %
                          (error_score, format_exc()),
                          FitFailedWarning)
        else:
            raise ValueError("error_score must be the string 'raise' or a"
                             " numeric value. (Hint: if using 'raise', please"
                             " make sure that it has been spelled correctly.)")
    
    
    else:
        # The estimator is fitted to data correctly
        # Calculate scoring of estimator using custom scorer
        fit_time = time.time() - start_time
        test_scores = _score(estimator, X_test, y_test, scorer)
        score_time = time.time() - start_time - fit_time
        if return_train_score:
            train_scores = _score(estimator, X_train, y_train, scorer)

    if verbose > 1:
        total_time = score_time + fit_time
        end_msg = f"[CV{progress_msg}] END "
        result_msg = params_msg + (";" if params_msg else "")
        if verbose > 2:
            if isinstance(test_scores, dict):
                for scorer_name in sorted(test_scores):
                    result_msg += f" {scorer_name}: ("
                    if return_train_score:
                        result_msg += (f"train="
                                       f"{train_scores[scorer_name]:.3f}, ")
                    result_msg += f"test={test_scores[scorer_name]:.3f})"
        result_msg += f" total time={logger.short_format_time(total_time)}"

        # Right align the result_msg
        end_msg += "." * (80 - len(end_msg) - len(result_msg))
        end_msg += result_msg
        print(end_msg)

    result["test_scores"] = test_scores
    if return_train_score:
        result["train_scores"] = train_scores
    if return_n_test_samples:
        # Return the number of bags
        result["n_test_samples"] = _num_samples(X_test)
    if return_times:
        result["fit_time"] = fit_time
        result["score_time"] = score_time
    if return_parameters:
        result["parameters"] = parameters
    if return_estimator:
        result["estimator"] = estimator
        
    return result


def _BagScorer_estimator_fit(estimator, X_train, y_train, scorer, **fit_params):
    
    generic_error = ("Invalid scorer object. Should be BagScorer or dictionary"
                     " with structure {'key':BagScorer} object. Got {scorer!r}")
    
    if y_train is None:
        # X_train is an iterabls of bags NOT single-instance examples
        if isinstance(scorer, BagScorer):
            # Depreciated
            estimator = scorer.estimator_fit(estimator, 
                                             X_train,
                                             y_train=None, 
                                             **fit_params)
        elif isinstance(scorer, dict):
            # It is possible to pass a dictionary of scorers
            for metric, bagScorer in scorer.items():
                if isinstance(bagScorer, BagScorer):
                    estimator = bagScorer.estimator_fit(estimator, 
                                                        X_train, 
                                                        y_train=None, 
                                                        **fit_params)
        else:
            raise ValueError(generic_error)
        
    else:
        # X_train, y_train are iterabls of bags NOT single-instance examples
        if isinstance(scorer, BagScorer):
            # Depreciated
            estimator = scorer.estimator_fit(estimator, 
                                             X_train,
                                             y_train, 
                                             **fit_params)
        elif isinstance(scorer, dict):
            # It is possible to pass a dictionary of scorers
            for metric, bagScorer in scorer.items():
                if isinstance(bagScorer, BagScorer):
                    estimator = bagScorer.estimator_fit(estimator, 
                                                        X_train, 
                                                        y_train, 
                                                        **fit_params)
        else:
            raise ValueError(generic_error)
        
    return estimator


@_deprecate_positional_args
def cross_validate_bag(estimator, X, y=None, *, groups=None, scoring=None, cv=None,
                   n_jobs=None, verbose=0, fit_params=None,
                   pre_dispatch='2*n_jobs', return_train_score=False,
                   return_estimator=False, error_score=np.nan):
    """Evaluate metric(s) by cross-validation and also record fit/score times.
    Read more in the:ref:`User Guide <multimetric_cross_validation>`.
    Parameters
    ----------
    estimator: estimator object implementing 'fit'
        The object to use to fit the data.
    X: array-like of shape (n_samples, n_features)
        The data to fit. Can be for example a list, or an array.
    y: array-like of shape (n_samples,) or (n_samples, n_outputs), \
            default=None
        The target variable to try to predict in the case of
        supervised learning.
    groups: array-like of shape (n_samples,), default=None
        Group labels for the samples used while splitting the dataset into
        train/test set. Only used in conjunction with a "Group":term:`cv`
        instance (e.g.,:class:`GroupKFold`).
    scoring: str, callable, list/tuple, or dict, default=None
        A single str (see :ref:`scoring_parameter`) or a callable
        (see :ref:`scoring`) to evaluate the predictions on the test set.
        For evaluating multiple metrics, either give a list of (unique) strings
        or a dict with names as keys and callables as values.
        NOTE that when using custom scorers, each scorer should return a single
        value. Metric functions returning a list/array of values can be wrapped
        into multiple scorers that return one value each.
        See :ref:`multimetric_grid_search` for an example.
        If None, the estimator's score method is used.
    cv: int, cross-validation generator or an iterable, default=None
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:
        - None, to use the default 5-fold cross validation,
        - int, to specify the number of folds in a `(Stratified)KFold`,
        - :term:`CV splitter`,
        - An iterable yielding (train, test) splits as arrays of indices.
        For int/None inputs, if the estimator is a classifier and ``y`` is
        either binary or multiclass, :class:`StratifiedKFold` is used. In all
        other cases, :class:`KFold` is used.
        Refer :ref:`User Guide <cross_validation>` for the various
        cross-validation strategies that can be used here.
        .. versionchanged:: 0.22
            ``cv`` default value if None changed from 3-fold to 5-fold.
    n_jobs: int, default=None
        The number of CPUs to use to do the computation.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
        for more details.
    verbose: int, default=0
        The verbosity level.
    fit_params: dict, default=None
        Parameters to pass to the fit method of the estimator.
    pre_dispatch: int or str, default='2*n_jobs'
        Controls the number of jobs that get dispatched during parallel
        execution. Reducing this number can be useful to avoid an
        explosion of memory consumption when more jobs get dispatched
        than CPUs can process. This parameter can be:
            - None, in which case all the jobs are immediately
              created and spawned. Use this for lightweight and
              fast-running jobs, to avoid delays due to on-demand
              spawning of the jobs
            - An int, giving the exact number of total jobs that are
              spawned
            - A str, giving an expression as a function of n_jobs,
              as in '2*n_jobs'
    return_train_score: bool, default=False
        Whether to include train scores.
        Computing training scores is used to get insights on how different
        parameter settings impact the overfitting/underfitting trade-off.
        However computing the scores on the training set can be computationally
        expensive and is not strictly required to select the parameters that
        yield the best generalization performance.
        .. versionadded:: 0.19
        .. versionchanged:: 0.21
            Default value was changed from ``True`` to ``False``
    return_estimator: bool, default=False
        Whether to return the estimators fitted on each split.
        .. versionadded:: 0.20
    error_score: 'raise' or numeric
        Value to assign to the score if an error occurs in estimator fitting.
        If set to 'raise', the error is raised.
        If a numeric value is given, FitFailedWarning is raised. This parameter
        does not affect the refit step, which will always raise the error.
        .. versionadded:: 0.20
    Returns
    -------
    scores: dict of float arrays of shape (n_splits,)
        Array of scores of the estimator for each run of the cross validation.
        A dict of arrays containing the score/time arrays for each scorer is
        returned. The possible keys for this ``dict`` are:
            ``test_score``
                The score array for test scores on each cv split.
                Suffix ``_score`` in ``test_score`` changes to a specific
                metric like ``test_r2`` or ``test_auc`` if there are
                multiple scoring metrics in the scoring parameter.
            ``train_score``
                The score array for train scores on each cv split.
                Suffix ``_score`` in ``train_score`` changes to a specific
                metric like ``train_r2`` or ``train_auc`` if there are
                multiple scoring metrics in the scoring parameter.
                This is available only if ``return_train_score`` parameter
                is ``True``.
            ``fit_time``
                The time for fitting the estimator on the train
                set for each cv split.
            ``score_time``
                The time for scoring the estimator on the test set for each
                cv split. (Note time for scoring on the train set is not
                included even if ``return_train_score`` is set to ``True``
            ``estimator``
                The estimator objects for each cv split.
                This is available only if ``return_estimator`` parameter
                is set to ``True``.
    Examples
    --------
    >>> from sklearn import datasets, linear_model
    >>> from sklearn.model_selection import cross_validate
    >>> from sklearn.metrics import make_scorer
    >>> from sklearn.metrics import confusion_matrix
    >>> from sklearn.svm import LinearSVC
    >>> diabetes = datasets.load_diabetes()
    >>> X = diabetes.data[:150]
    >>> y = diabetes.target[:150]
    >>> lasso = linear_model.Lasso()
    Single metric evaluation using ``cross_validate``
    >>> cv_results = cross_validate(lasso, X, y, cv=3)
    >>> sorted(cv_results.keys())
    ['fit_time', 'score_time', 'test_score']
    >>> cv_results['test_score']
    array([0.33150734, 0.08022311, 0.03531764])
    Multiple metric evaluation using ``cross_validate``
    (please refer the ``scoring`` parameter doc for more information)
    >>> scores = cross_validate(lasso, X, y, cv=3,
    ...                         scoring=('r2', 'neg_mean_squared_error'),
    ...                         return_train_score=True)
    >>> print(scores['test_neg_mean_squared_error'])
    [-3635.5... -3573.3... -6114.7...]
    >>> print(scores['train_r2'])
    [0.28010158 0.39088426 0.22784852]
    See Also
    ---------
    :func:`sklearn.model_selection.cross_val_score`:
        Run cross-validation for single metric evaluation.
    :func:`sklearn.model_selection.cross_val_predict`:
        Get predictions from each split of cross-validation for diagnostic
        purposes.
    :func:`sklearn.metrics.make_scorer`:
        Make a scorer from a performance metric or loss function.
    """
    X, y, groups = indexable(X, y, groups)

    cv = check_cv(cv, y, classifier=is_classifier(estimator))
    scorers = _check_multimetric_scoring(estimator, scoring=scoring)

    # We clone the estimator to make sure that all the folds are
    # independent, and that it is pickle-able.
    parallel = Parallel(n_jobs=n_jobs, verbose=verbose,
                        pre_dispatch=pre_dispatch)
    results = parallel(
        delayed(_fit_and_score)(
            clone(estimator), X, y, scorers, train, test, verbose, None,
            fit_params, return_train_score=return_train_score,
            return_times=True, return_estimator=return_estimator,
            error_score=error_score)
        for train, test in cv.split(X, y, groups))

    # results is a list of dictonaries
    results = _aggregate_score_dicts(results)
    ret = {}
    ret['fit_time'] = results['fit_time']
    ret['score_time'] = results['score_time']

    # results['test_scores'] is a list of dictionaries
    test_scores = _aggregate_score_dicts(results['test_scores'])
    if return_train_score:
        train_scores = _aggregate_score_dicts(results['train_scores'])

    if return_estimator:
        fitted_estimators = results['estimator']
        ret['estimator'] = fitted_estimators

    for name in scorers:
        ret['test_%s' % name] = test_scores[name]
        if return_train_score:
            key = 'train_%s' % name
            ret[key] = train_scores[name]

    return ret



def bags_2_si_generator(bags, bag_labels, sparse_input=False):
    """Convert a n x (m x p) array of bag instances into a k x p array of
    instances. n is the number of bags, and m is the number of instances within
    each bag. m can vary per bag. k is the total number of instances within
    all bags. k = sum (m for bag in n). p is the feature space of each instance
    inputs
    -------
    bags: (iterable) containing bags of shape (m x p) sparse arrays
    bag_labels: (iterable) containing labels assocaited with each bag. Labels
        are expanded and each instance within a bag inherits the label of the
        bag
    sparse_input: (bool) if True, the output instances are left as a sparse array.
        Some sklearn estimators can handle sparse feature inputs
    output
    -------
    instances, labels: (generator) """

    for bag, label in zip(bags, bag_labels):
        # Unpack bag into instances

        if sparse_input and isinstance(bag, csr_matrix):
            # Keep instances as sparse array
            instances = bag # Sparse array
        elif sparse_input and not isinstance(bag, csr_matrix):
            # Convert instances to csr matrix if sparse=True and the bag
            # is not currently a csr matrix
            instances = csr_matrix(bag) # Dense array
        elif not sparse_input and isinstance(bag, csr_matrix):
            # Convert csr_array to dense array
            instances = bag.toarray()
        elif not sparse_input and not isinstance(bag, csr_matrix):
            # Keep as dense array
            instances = bag # Dense array
        else:
            msg=('bags must be sparse array (scipy.sparse.csr_matrix) or' +
            ' dense array (np.array). Got type {}')
            raise TypeError(msg.format(type(bag)))

        labels = np.array([label].__mul__(instances.shape[0]))

        yield instances, labels


def bags_2_si( bags, bag_labels, sparse_input=False):
    """Convert a n x (m x p) array of bag instances into a k x p array of
    instances. n is the number of bags, and m is the number of instances within
    each bag. m can vary per bag. k is the total number of instances within
    all bags. k = sum (m for bag in n). p is the feature space of each instance
    inputs
    -------
    bags: (iterable) containing bags of shape (m x p) sparse arrays
    bag_labels: (iterable) containing labels assocaited with each bag. Labels
        are expanded and each instance within a bag inherits the label of the
        bag
    sparse: (bool) if True, the output instances are left as a sparse array.
        Some sklearn estimators can handle sparse feature inputs
    output
    -------
    instances, labels: (np.array) or (scipy.sparse.csr.csr_matrix)
    depending on 'sparse'"""

    # Initialize generator over bags
    bag_iterator = bags_2_si_generator(bags,
                                       bag_labels,
                                       sparse_input=sparse_input)

    # Initialize datasets
    instances, labels = [], []

    # Gather datasets
    for part_instances, part_labels in bag_iterator:

        instances.append(part_instances)
        labels.append(part_labels)

    # Flatten into otuput shape - [k x p] instances and [k] labels
    if sparse_input:
        # Row-stack sparse arrays into a sinlge  k x p sparse array
        instances = vstack(instances)
        labels = np.concatenate(labels)
    else:
        # Row-concatenate dense arrays into a single k x p array
        instances = np.concatenate(instances)
        labels = np.concatenate(labels)

    return instances, labels



class BagScorer:
    """This is a custom scoring object for use with sklearn cross validation
    model evaluation. This includes cross_validate, cross_val_score, and
    GridSearchCV

    This scoring object is specifically used for scoring bag labels predicted
    from single-instance predictions within the bag

    According to sklearn, this scorer must be
    1. It can be called with parameters (estimator, X, y), where estimator is
    the model that should be evaluated, X is validation data, and y is
    the ground truth target for X
    2. Return a floating point number that quantifies the estimator prediction
    quality on X with reference to y. If the metric is a loss then the value
    should be negated (more positive is better)

    """
    def __init__(self, scorer, sparse_input=False):
        """This class should NOT be passed directly as the 'scorer' argument
        to cross_validate, cross_val_score, or GridSearchCV without
        initializing the class
        
        When scorer is a dictionary of values, raise an error
        When the scorer is a callable with method _score_func then return
            the score from this scoring metric
        
        inputs
        ------
        scorer: (#TODO)
        sparse_input: (bool) if True, the output instances are left as a sparse array.
        Some sklearn estimators can handle sparse feature inputs"""

        # Initialization parameters
        self.sparse_input = sparse_input

        # Metric for scoring bags
        if isinstance(scorer, dict):
            msg='scorer cannot be type dict. Passed type {}'
            raise ValueError(msg.format(type(scorer)))

        if not hasattr(scorer, '_score_func'):
            msg=('scorer object must have callable "_score_func" with a'+
            ' signature f(y_true, y_pred)')
            raise ValueError(msg)

        self.scorer = scorer

        return None


    def __call__(self, estimator, *positional_args, **kwargs): # TODO This should accept *args, **kwargs
        """
        inputs
        -------
        estimator: () the model that should be evaluated
        X: (iterable) is validation data. It has to be an iterable of bags,
        for example an [n x (m x p)] array of bag instances. n is the number
        of bags, and m is the number of instances within each bag.
        p is the feature space of each instance
        y: () is the ground truth target for X
        outputs
        -------
        score: (float) result of score of estimator on bags
        """
        # Initialize positional and keyword arguments
        X = positional_args[0] # Validation data (for predictions)
        y = positional_args[1] # Ground truth (for metrics)

        # Test if estimator is fitted already or not
        check_is_fitted(estimator)

        # Predict on bags
        bag_predictions = self.predict_bags(estimator, X)

        # Calculate metrics - API call to scorer function
        if hasattr(self.scorer, '_score_func'):
            kwargs = self.scorer._kwargs
            ret = self.scorer._score_func(y, bag_predictions, **kwargs)

        else:
            msg=('scorer object must have callable "_score_func" with a'+
            ' signature f(y_true, y_pred)')
            raise ValueError(msg)

        return ret
    

    def estimator_fit(self, estimator, X_train, y_train=None, **fit_params):
        """The sklearn _fit_and_score method requires estimator fitting to be
        done as a separate task from scoring
        inputs
        -------
        estimator: ()
        X_train: (iterable) is validation data. It has to be an iterable of bags,
        for example an [n x (m x p)] array of bag instances. n is the number
        of bags, and m is the number of instances within each bag.
        y_train: () is the ground truth target for X
        **fit_params: ()
        outputs
        ------
        estimator: fitted estimator"""

        if y_train is None:
            msg=('Single Instance labels cannot be constructed from bags if' +
            ' y_train is None. Pass bag labels to validation')
            raise ValueError(msg)

        # Find SI data
        SI_examples, SI_labels = bags_2_si(X_train, y_train, self.sparse)

        if y_train is None:
            # This should not happen - see ValueError
            estimator.fit(SI_examples, **fit_params)
        else:
            estimator.fit(SI_examples, SI_labels, **fit_params)

        return estimator # Fitted estimator

    @staticmethod
    def reduce_bag_label(predictions, method: str = 'mode') -> Union[str, int]:
        """Determine the bag label from the single-instance classifications of 
        its members. 'mode' returns the most frequently occuring bag label
        inputs
        -------
        predictions: (iterable) of labels
        method: (str) 'mode' is only supported.
            TODO support based on label weights
        outputs
        -------
        label: (str / int) of most common prediction"""

        if method == 'mode':
            label, count = Counter(predictions).most_common(1)[0]
        else:
            method_error = "only 'mode' is supported as a reduction method." \
                "Got {}".format(method)
            raise ValueError(method_error)

        return label

    @classmethod
    def predict_bags(cls, estimator, bags, method: str ='mode'):
        """
        inputs
        ------
        estimator: ()
        bags: ()
        method: (string)
        outputs
        -------"""

        bag_predictions = []

        for bag in bags:
            # Predict labels in bag
            si_predictions = estimator.predict(bag)
            bag_prediction = cls.reduce_bag_label(si_predictions, method=method)
            bag_predictions.append(bag_prediction)

        return bag_predictions



from .bag_cross_validate import (cross_validate_bag, 
    bags_2_si, 
    bags_2_si_generator, 
    BagScorer,
    _fit_and_score)


__author__ = 'John Vorsten'
__email__ = 'vorstenjohn@gmail.com'
__version__ = '0.0.1'



# -*- coding: utf-8 -*-
"""
Created on Sat Jun 27 15:18:53 2020

@author: z003vrzk
"""

# Python imports
import sys, os
import unittest

# Sklearn imports
from sklearn.naive_bayes import ComplementNB
from sklearn.dummy import DummyClassifier
from sklearn.metrics import accuracy_score, make_scorer, precision_score, recall_score
from sklearn.model_selection import ShuffleSplit, KFold
from sklearn.metrics._scorer import _check_multimetric_scoring
from sklearn.base import is_classifier, clone
from sklearn.metrics._scorer import check_scoring
from sklearn.model_selection._split import check_cv
from sklearn.utils import indexable
from sklearn.neighbors import KNeighborsClassifier

# Third party imports
from joblib import Parallel, delayed
import numpy as np
from scipy.sparse import csr_matrix

# Local imports
from bag_cross_validate import (BagScorer, 
                                cross_validate_bag,
                                _fit_and_score,
                                bags_2_si,)


#%%

class TestBagScorer(unittest.TestCase):
    
    def setUp(self):
        
        """Generate some dummy data
        Create bags and single-instance data
        A set of bags have a shape [n x (m x p)], and can be through of as an
        array of bag instances.
        n is the number of bags
        m is the number of instances within each bag (this can vary between bags)
        p is the feature space of each instance"""
        n_bags = 100
        m_instances = 5 # Static number of bags
        p = 5
        bags = []
        # 25% negative class, 75% positive class
        # Bags are created with random data, of shape (n, (m,p))
        labels = np.concatenate((np.ones(int(n_bags*0.5)),
                                 np.zeros(int(n_bags*(1-0.5))),
                                 ))
        for _ in range(n_bags):
            _rand = np.random.rand(m_instances, p)
            bag = np.where(_rand < 0.25, 1, 0)
            bags.append(bag)
        bags = np.array(bags)

        # Split dummy dataset dataset
        rs = ShuffleSplit(n_splits=1, test_size=0.2, train_size=0.8)
        train_index, test_index = next(rs.split(bags, labels))
        train_bags, train_labels = bags[train_index], labels[train_index]
        test_bags, test_labels = bags[test_index], labels[test_index]
        self.train_bags, self.train_labels = bags[train_index], labels[train_index]
        self.test_bags, self.test_labels = bags[test_index], labels[test_index]
        
        return None


    def test_BagScorer(self):

        """Define scoring functions, such as accuracy or recall,
        which will be used to score how well single-instance inference
        performs on the bag classification task

        The scoring functions have some requirements -
        a) They are passed to BagScorer on initialization
        b) Must have a method "_score_func" with a signature f(y_true, y_pred)
            (This is provided by default when using sklearn.metrics.make_scorer)

        """
        
        # Create scoring metrics, and load scoring metric into BagScorer
        accuracy_scorer = make_scorer(accuracy_score, normalize=True)
        precision_scorer = make_scorer(precision_score, average='weighted')
        recall_scorer = make_scorer(recall_score, average='weighted')
        # {'normalize':'weighted'}
        self.assertDictContainsSubset({'normalize':True}, accuracy_scorer._kwargs)
        self.assertIn('_score_func', accuracy_scorer.__dict__.keys())

        # Dummy data
        train_bags, train_labels = self.train_bags, self.train_labels
        test_bags, test_labels = self.test_bags, self.test_labels

        # Create a single-instance estimator
        compNB = ComplementNB(alpha=1.0, fit_prior=True, class_prior=None, norm=False)

        # Test custom scorer
        bagAccScorer = BagScorer(accuracy_scorer, sparse=True)
        bagPrecisionScorer = BagScorer(precision_scorer, sparse=True)
        bagRecallScorer = BagScorer(recall_scorer, sparse=True)
        estimator = bagAccScorer.estimator_fit(compNB, train_bags, train_labels)
        
        # The estimator is the same for all instances...
        accuracy = bagAccScorer(estimator, test_bags, test_labels)
        precision = bagPrecisionScorer(estimator, test_bags, test_labels)
        recall = bagRecallScorer(estimator, test_bags, test_labels)

        self.assertIsInstance(accuracy, float)
        self.assertLess(accuracy, 1)
        self.assertGreater(accuracy, 0)
        
        self.assertIsInstance(precision, float)
        self.assertLess(precision, 1)
        self.assertGreater(precision, 0)
        
        self.assertIsInstance(recall, float)
        self.assertLess(recall, 1)
        self.assertGreater(recall, 0)
        
        return None
    

    def test_scorer_signature(self):

        """Define scoring functions, such as accuracy or recall,
        which will be used to score how well single-instance inference
        performs on the bag classification task

        The scoring functions have some requirements -
        a) They are passed to BagScorer on initialization
        b) Must have a method "_score_func" with a signature f(y_true, y_pred)
            (This is provided by default when using sklearn.metrics.make_scorer)

        """
        accuracy_scorer = make_scorer(accuracy_score, normalize='weighted')
        print(accuracy_scorer._kwargs) # {'normalize':'weighted'}
        hasattr(accuracy_scorer, '_score_func') # True

        self.assertTrue(hasattr(accuracy_scorer, '_score_func'))


    def test_BagScorer_signature(self):

        # Test custom scorer
        accuracy_scorer = make_scorer(accuracy_score, normalize='weighted')
        bagAccScorer = BagScorer(accuracy_scorer, sparse=True)

        self.assertTrue(callable(bagAccScorer), msg="BagScorer must be callable")
        
        return None


    def test_BagScorer_metric(self):

        """Define scoring functions, such as accuracy or recall,
        which will be used to score how well single-instance inference
        performs on the bag classification task

        The scoring functions have some requirements -
        a) They are passed to BagScorer on initialization
        b) Must have a method "_score_func" with a signature f(y_true, y_pred)
            (This is provided by default when using sklearn.metrics.make_scorer)

        Successful conditions:
            The bagscorer must report the same performance metrics as when the 
            metrics are manually calculated
        This tests if the bagscorer property fits, trains, and evaluates
        the estimator passed to it
        """
        
        # Generate a scoring metric for the bag scorer
        accuracy_scorer = make_scorer(accuracy_score)
        self.assertTrue(hasattr(accuracy_scorer, '_score_func'), 
                        msg='accuracy scorer must have _score_function method') 

        # Generate some data
        train_bags, train_labels = self.train_bags, self.train_labels
        test_bags, test_labels = self.test_bags, self.test_labels

        # Create a dummy estimator
        dumb = DummyClassifier(strategy='constant', constant=1)
        
        # concatenate arrays across 1st axis
        SI_train, SI_train_labels = bags_2_si(train_bags, train_labels)
        SI_test, SI_test_labels = bags_2_si(test_bags, test_labels)
        dumb.fit(SI_train, SI_train_labels)
        pred_test = dumb.predict(SI_test)
        pred_train = dumb.predict(SI_train)

        """Calculate the correct number of predictions based on dummy classifier
        The dummy classifier predicts 1 always (constant)
        The training set bas """
        pct_train = sum(train_labels) / len(train_labels)
        pct_test = sum(test_labels) / len(test_labels)
        dumb_accuracy_train = accuracy_score(SI_train_labels, pred_train)
        dumb_accuracy_test = accuracy_score(SI_test_labels, pred_test)

        # Test custom scorer, with the same dummy estimator
        bagAccScorer = BagScorer(accuracy_scorer, sparse=True)
        estimator = bagAccScorer.estimator_fit(dumb, train_bags, train_labels)
        test_score = bagAccScorer(estimator, test_bags, test_labels)
        train_score = bagAccScorer(estimator, train_bags, train_labels)

        """test_score should output the accuracy for predictions among bags
        The test_score for bagScorer should be equal to the dumb_accuracy_test
        because bag labels are reduced by the most frequest SI prediction

        If all SI labels are predicted + then all bags will be predicted +
        The accuracy of bag labels reduced by BagScorer will be equal to
        percent of bag labels that are positive"""

        self.assertEqual(test_score, pct_test)
        self.assertEqual(train_score, pct_train)
        self.assertEqual(pct_train, dumb_accuracy_train)
        self.assertEqual(pct_test, dumb_accuracy_test)


    def test_cross_validate_bag(self):

        # Scoring
        accuracy_scorer = make_scorer(accuracy_score, normalize='weighted')

        # Dummy data
        train_bags, train_labels = self.train_bags, self.train_labels
        test_bags, test_labels = self.test_bags, self.test_labels

        # Define an estimator
        dumb = DummyClassifier(strategy='constant', constant=1)
        
        # Calculate metrics manually
        expected_accuracy = sum(train_labels) / len(train_labels)
        kf = KFold(n_splits = 4)
        accuracies = []
        for train_index, test_index in kf.split(train_labels):
            _fold = train_labels[test_index]
            _acc = sum(_fold) / len(_fold)
            print(sum(_fold))
            accuracies.append(_acc)
        print('Global Accuracy : ', sum(train_labels) / len(train_labels))
        print('Averaged accuracies : ', np.mean(accuracies))

        # Custom scorer
        bagAccScorer = BagScorer(accuracy_scorer, sparse=True)
        scorer = {'bag-accuracy-scorer': bagAccScorer,
                   }

        # Test cross_validate_bag
        # Res is a dictonary of lists {'fit_time':[1,2,3],
        # 'test_bag-accuracy-scorer':[0.1,0.2,0.3]}
        res = cross_validate_bag(dumb, train_bags, train_labels,
                             cv=4, scoring=scorer,
                             n_jobs=1, verbose=0, fit_params=None,
                             pre_dispatch='2*n_jobs', return_train_score=False,
                             return_estimator=False, error_score='raise')

        """The arithmetic mean of all accuracy predictions should equal the
        prediction accuracy of the training bags (At least if all splits are
        equal size -> Which is not true if the number of training instances
        is not divisible by the number of splits)
        This is only true because the dummy classifier always predicts 1
        If the splits are not equal size then they will be close to equal"""
        self.assertAlmostEqual(np.mean(res['test_bag-accuracy-scorer']), 
                               expected_accuracy, 3)
        # Just check the mean also LOL
        self.assertEqual(np.mean(res['test_bag-accuracy-scorer']), 
                         expected_accuracy)
        # 4 Crossvalidation splits
        self.assertTrue(len(res['test_bag-accuracy-scorer']) == 4)
        # Assert result has dictionary values
        self.assertIn('fit_time', res.keys())
        self.assertIn('score_time', res.keys())
        
        return None


    def test_fit_and_score(self):

        # Scoring
        accuracy_scorer = make_scorer(accuracy_score, normalize='weighted')

        # Test estimator
        dumb = DummyClassifier(strategy='constant', constant=1)

        # Test custom scorer
        bagAccScorer = BagScorer(accuracy_scorer, sparse=True)

        # _fit_and_score testing
        X = self.train_bags
        y = self.train_labels
        scoring = {'bag-accuracy-scorer': bagAccScorer,
                   }
        estimator = dumb
        groups = None
        cv = 3
        n_jobs=3
        verbose=0
        pre_dispatch=6
        fit_params=None
        return_estimator=None
        error_score='raise'
        return_train_score=None
        parameters=None

        # Test _fit_and_score method
        X, y, groups = indexable(X, y, groups)
        cv = check_cv(cv, y, classifier=is_classifier(estimator))
        scorers = _check_multimetric_scoring(estimator, scoring=scoring)

        # We clone the estimator to make sure that all the folds are
        # independent, and that it is pickle-able.
        parallel = Parallel(n_jobs=n_jobs, verbose=verbose,
                            pre_dispatch=pre_dispatch)
        
        # Scores is a list of dictonaries
        """When scoring is a dictionary, the returned result looks like
        [{'test_scores': {'bag-accuracy-scorer': 0.5185185185185185},
          'fit_time': 0.0,
          'score_time': 0.0},
         {'test_scores': {'bag-accuracy-scorer': 0.5185185185185185},
          'fit_time': 0.0,
          'score_time': 0.0}, ... ]"""
        scores = parallel(
            delayed(_fit_and_score)(
                clone(estimator), X, y, scorers, train, test, verbose, parameters,
                fit_params, return_train_score=return_train_score,
                return_times=True, return_estimator=return_estimator,
                error_score=error_score)
            for train, test in cv.split(X, y, groups))
        
        for score in scores:
            bag_scoring_metric = score['test_scores']
            self.assertLessEqual(bag_scoring_metric['bag-accuracy-scorer'], 1)
            self.assertGreaterEqual(bag_scoring_metric['bag-accuracy-scorer'], 0)
            
            fit_time = score['fit_time']
            self.assertIsInstance(fit_time, float)
            
            score_time = score['score_time']
            self.assertIsInstance(score_time, float)
        
        return None
    
    
    def test_fit_and_score_return_dict(self):
        
        # Scoring
        accuracy_scorer = make_scorer(accuracy_score, normalize='weighted')
        
        # Test estimator
        dumb = DummyClassifier(strategy='constant', constant=1)

        # Test custom scorer
        bagAccScorer = BagScorer(accuracy_scorer, sparse=True)
        
        # Rename for easier parameters
        X = self.train_bags
        y = self.train_labels
        scoring = {'bag-scorer':bagAccScorer}
        estimator = dumb
        groups = None
        cv = 3
        n_jobs=3
        verbose=0
        pre_dispatch=6
        fit_params=None
        return_estimator=True
        error_score='raise'
        return_train_score=True
        parameters=None
        
        # Test _fit_and_score method
        X, y, groups = indexable(X, y, groups)
        cv = check_cv(cv, y, classifier=is_classifier(estimator))
        scorers = _check_multimetric_scoring(estimator, scoring=scoring)

        # Use one cross-validation split
        generator = cv.split(X, y, groups)
        # Get training and test split of training data
        train, test = next(generator)
        # Generate scores using BagScorer
        scores = _fit_and_score(
            clone(estimator), X, y, scorers, train, test, verbose, parameters,
            fit_params, 
            return_train_score=return_train_score,
            return_times=True, 
            return_estimator=return_estimator,
            return_n_test_samples=False,
            error_score=error_score)
        
        # Returned dictionary contains keys
        self.assertIn('train_scores', scores.keys())
        self.assertIn('test_scores', scores.keys())
        self.assertIn('fit_time', scores.keys())
        self.assertIn('score_time', scores.keys())
        self.assertIn('estimator', scores.keys())
        
        return None



class TestCrossValidation(unittest.TestCase):
    
    def setUp(self):
        """Generate some dummy data
        Create bags and single-instance data
        A set of bags have a shape [n x (m x p)], and can be through of as an
        array of bag instances.
        n is the number of bags
        m is the number of instances within each bag (this can vary between bags)
        p is the feature space of each instance"""
        n_bags = 100
        m_instances_range = [5,10] # Dynamic number instance per bag
        p = 5
        bags_sparse = []
        # 25% negative class, 75% positive class
        # Bags are created with random data, of shape (n, (m,p))
        labels = np.concatenate((np.ones(int(n_bags*0.5)),
                                 np.zeros(int(n_bags*(1-0.5))),
                                 ))
        for n in range(0, n_bags):
            m_instances = np.random.randint(m_instances_range[0], 
                                            m_instances_range[1],)
            bag = np.random.randint(low=0, high=2, size=(m_instances, p))
            bag_sparse = csr_matrix(bag)
            bags_sparse.append(bag_sparse)
            
        bags_sparse = np.array(bags_sparse)

        # Split dummy dataset dataset
        rs = ShuffleSplit(n_splits=1, test_size=0.2, train_size=0.8)
        train_index, test_index = next(rs.split(bags_sparse, labels))
        train_bags, train_labels = bags_sparse[train_index], labels[train_index]
        test_bags, test_labels = bags_sparse[test_index], labels[test_index]
        self.train_bags, self.train_labels = bags_sparse[train_index], labels[train_index]
        self.test_bags, self.test_labels = bags_sparse[test_index], labels[test_index]
        
        return None
    
    
    def test_bag_cross_validate_sparse(self):
        
        # Scoring
        accuracy_scorer = make_scorer(accuracy_score, normalize='weighted')

        # Dummy data
        train_bags, train_labels = self.train_bags, self.train_labels
        test_bags, test_labels = self.test_bags, self.test_labels

        # Define an estimator
        estimator = KNeighborsClassifier(n_neighbors=10, 
                                         weights='uniform',
                                         algorithm='ball_tree', 
                                         n_jobs=4)

        # Custom scorer
        bagAccScorer = BagScorer(accuracy_scorer, input_sparse=True)
        scorer = {'bag-accuracy-scorer': bagAccScorer,
                  }

        # Test cross_validate_bag
        # Res is a dictonary of lists {'fit_time':[1,2,3],
        # 'test_bag-accuracy-scorer':[0.1,0.2,0.3]}
        res = cross_validate_bag(estimator, train_bags, train_labels,
                                 cv=4, scoring=scorer,
                                 n_jobs=1, verbose=0, fit_params=None,
                                 pre_dispatch='2*n_jobs', return_train_score=False,
                                 return_estimator=False, error_score='raise')


        # 4 Crossvalidation splits
        self.assertTrue(len(res['test_bag-accuracy-scorer']) == 4)
        # Assert result has dictionary values
        self.assertIn('fit_time', res.keys())
        self.assertIn('score_time', res.keys())
        
        return None












#%%

if __name__ == '__main__':
    # Run all test cases
    unittest.main()
    
    # # Run specific test methods...
    # runner = unittest.TextTestRunner()
    # classes = [TestBagScorer]
    # unit_tests_to_run = [
    #     'test_BagScorer',
    #     'test_BagScorer_dict',
    #     'test_BagScorer_metric',
    #     'test_BagScorer_signature',
    #     'test_cross_validate_bag',
    #     'test_fit_and_score',
    #     'test_fit_and_score_return_dict',
    #     'test_scorer_signature',
    #     ]
    
    # Run specific test methods... (altenative method)
    # suite = unittest.TestSuite()
    # suite.addTest(TestBagScorer('test_fit_and_score_return_dict'))
    # runner = unittest.TextTestRunner()
    # runner.run(suite)